/* dcf_animation.c
 from dc fighting v0.2.0
  by
   jimbo
    Dan Potter and Jordan DeLong of Cryptic Allusion
     Andrew K.
      Marcus Comstedt


animation functions:
 void add_anchor( fighter_t *p, double x, double *y, int state );
 void add_partial_anchor( fighter_t *p, double x, int which_params, double *y, int state );
 void animation_timestep( fighter_t *p );

depends on the SPACIAL_OFFSET #define
 partials will depend on different arm/leg starting/ending locations...
  this better be made modular enough to handle maximum models with minimum hacking of code..

v0.2.4: took out the SPACIAL_OFFSET code, working on partials...


 stick them in the regular list?  or a separate for arms & legs?




 */



/* the function version 
 *  v0.1.2 contains fighting state
 *
 */
void add_anchor( fighter_t *p, double x, double *y, int state ) {
  int i;
  /*** add current position as an anchor if list empty ***/
  if ( p->lists_in_future == 0 ) { 
    TIME_OF( 0 ) = 0.0; 
    PARAMS_OF( 0 ) = PARAMS_ALL; 

    for ( i=OFFSET_HACK; i< NUMBER_PARAMETERS + OFFSET_HACK; i++ ) { 
      p->future_lists[i][  (p->future_pointer + p->lists_in_future ) % FUTURE_SIZE ] 
	= p->params[ i-OFFSET_HACK ];
    } 
    p->lists_in_future++; 
  }
  /*** if buffer not full ***/
  if ( p->lists_in_future < FUTURE_SIZE ) { 
    TIME_OF( p->lists_in_future ) = x ;
    STATE_OF( p->lists_in_future ) = state ;
    PARAMS_OF( p->lists_in_future ) = PARAMS_ALL; 

    if ( TIME_OF( p->lists_in_future - 1) > 0.0 ) { 
      TIME_OF( p->lists_in_future ) += TIME_OF( p->lists_in_future - 1 ); 
    }

    for ( i=OFFSET_HACK; i< NUMBER_PARAMETERS + OFFSET_HACK; i++ ) { 
      p->future_lists[i][  (p->future_pointer + p->lists_in_future ) % FUTURE_SIZE ] 
	  = y [ p->reverse*NUMBER_PARAMETERS + i-OFFSET_HACK]; 
    } 
    p->lists_in_future++; 
  } 

}












#if 0

/* the function version 
 *  v0.1.8: switch on int which_params
 *
 *  - arms, legs, head, position....
 *
 * these #defines match up with those in the modular file fighter.c
 *
 *
 */
void add_partial_anchor( fighter_t *p, double x, int which_params, double *y, int state ) {
  int i;
  int i_begin, i_end;


  /* switch here : which partials does this anchor refer to?
   * - arms, legs
   * - position, rotation, both
   * - head, back...
   *
   *
   */
  switch( which_params ) {


  case PARAMS_SPATIAL:
    i_begin = POSITION_START;
    i_end = ROTATION_END;
    break;

  case PARAMS_LOCATION:
    i_begin = POSITION_START;
    i_end = POSITION_END;
    break;

  case PARAMS_ROTATION:
    i_begin = ROTATION_START;
    i_end = ROTATION_END;
    break;

  case PARAMS_ARMS:
    i_begin = ARMS_START;
    i_end = ARMS_END;
    break;

  case PARAMS_LEGS:
    i_begin = LEGS_START;
    i_end = LEGS_END;
    break;

  default: break;
  }



  /*** add current position as an anchor if list empty ***/
  if ( p->lists_in_future == 0 ) { 
    TIME_OF( 0 ) = 0.0;
    /*** v0.1.9: why not? 0+0=0 
     *  anchor all present configurations
     *  then only change the partials
     *
     *
     *  save for later
     *
     *
     */
    PARAMS_OF( 0 ) = which_params; 
    for ( i = OFFSET_HACK; i< NUMBER_PARAMETERS + OFFSET_HACK; i++ ) { 
      p->future_lists[i][  (p->future_pointer + p->lists_in_future ) % FUTURE_SIZE ] 
	= p->params[ i-OFFSET_HACK ];
    } 
    p->lists_in_future++; 
  }
  /*** if buffer not full ***/
  if ( p->lists_in_future < FUTURE_SIZE ) { 
    TIME_OF( p->lists_in_future ) = x ;
    STATE_OF( p->lists_in_future ) = state ;
    PARAMS_OF( p->lists_in_future ) = which_params; 

    if ( TIME_OF( p->lists_in_future - 1) > 0.0 ) { 
      TIME_OF( p->lists_in_future ) += TIME_OF( p->lists_in_future - 1 ); 
    } 
    for ( i = OFFSET_HACK + i_begin; i < i_end + OFFSET_HACK; i++ ) { 
      p->future_lists[i][  (p->future_pointer + p->lists_in_future ) % FUTURE_SIZE ] 
	  = y [ p->reverse*NUMBER_PARAMETERS + i-OFFSET_HACK]; 
    } 
    p->lists_in_future++; 
  } 

}



#endif
















/* v0.1.7: one more function to be replaced by a macro
 *
 *
 */
void animation_timestep( fighter_t *p ) {
  int i;
  int crossed = 0;

  
  /* Animation timestep accounting begins here:
   *
   * eliminate old anchors block - get rid of past events
   *  (cycle the buffer past them)
   */
  while (  ( TIME_OF( 0 )  < 0.0  )  &&
	   (  
             ( TIME_OF( 1 )  < 0.0 )
	       ||
	     (
	       ( TIME_OF(1) == 0.0 ) &&
	       ( TIME_OF(2) == 0.0 )
	     )
	   ) 
	) 
  {
    p->lists_in_future--;
    p->future_lists[0][ p->future_pointer++ ] = 0.0;
    p->future_pointer %= FUTURE_SIZE;
  }
  
  /*** advance the timestep of all pending anchors ***/
  for ( i=p->future_pointer; 
	i<p->future_pointer + p->lists_in_future;
	i++ ) {
    crossed = 0;
    /*** if it passes the 'now' line, apply the state ***/
    if ( ( p->future_lists[0][i % FUTURE_SIZE] < timestep )  &&
	 ( p->future_lists[0][i % FUTURE_SIZE] > 0.0 )  ) { 
      crossed = 1;
      
      /*  v0.1.7: do we want this? 
       *   - both of these lines are necessary for hits to take place...
       *      but they are undesirable, as the default state of a player
       *       should be discrete in time; as intermediate values for states
       *        will not properly convey range and type information embedded
       *         within...
       *
       */
      p->default_state = (int)p->future_lists[1][i % FUTURE_SIZE];
      p->fighting_state = p->default_state;
    }
    /*** timestep here ***/
    p->future_lists[0][i % FUTURE_SIZE] -= timestep;
  }
  
  /** find t parameter ***/
  t = TIME_OF( 1 ) == TIME_OF( 0 )
    ? 0.0 :
    ( -TIME_OF( 0 )  ) / (   TIME_OF( 1 ) - TIME_OF( 0 )   );
  
  /*** lock t into domain ***/
  if ( t < 0.0 ) t = 0.0;
  if ( t > 1.0 ) t = 0.0;
  
  /*** interpolate (offset for time in first slot) ***/
  if ( t != 0.0 ) {
    
    /* v0.1.8: ready for the partial interpolation of parameters...
     *   might not be necessary until later, if at all...
     *
     *   but here's the loop anyway.
     *
     *   assign the parameter range indicator to offset hack slot 2
     *   - only do those params
     */
    int i_begin, i_end;
    int which_params = PARAMS_OF( 1 );


#if 0
    /* switch here : which partials does this anchor refer to?
     * - arms, legs
     * - position, rotation, both
     * - head, back...
     *
     *
     */
    switch( which_params ) {
      
      
    case PARAMS_SPATIAL:
      i_begin = POSITION_START;
      i_end = ROTATION_END;
      break;
      
    case PARAMS_LOCATION:
      i_begin = POSITION_START;
      i_end = POSITION_END;
      break;
      
    case PARAMS_ROTATION:
      i_begin = ROTATION_START;
      i_end = ROTATION_END;
      break;
      
    case PARAMS_ARMS:
      i_begin = ARMS_START;
      i_end = ARMS_END;
      break;
      
    case PARAMS_LEGS:
      i_begin = LEGS_START;
      i_end = LEGS_END;
      break;
      
    case PARAMS_ALL:
    default: 
      i_begin = 3;
      i_end = NUMBER_PARAMETERS;
      break;
      
    }
    
#endif
    


    /* v0.2.1: new hack here - spatial offset
     *
     *
     *
     */

    i_begin = 0;
    i_end = NUMBER_PARAMETERS;

    for ( i= OFFSET_HACK + i_begin; i < i_end + OFFSET_HACK; i++ ) {
      
      /* hack v0.1.5 : dont interpolate the ROTATION_OPPONENT parameter
       *  maybe just a separate field in the struct for rotation...
       *   but then it wont be integrated into moves.
       *
       *
       * v0.2.2: getting rid of this hack wwith a separate position field in fightert_t struct
       *
       *
       */
      p->params[i-OFFSET_HACK] = 
	(1-t) * p->future_lists[i][p->future_pointer % FUTURE_SIZE]  + 
	t     * p->future_lists[i][(p->future_pointer+1) % FUTURE_SIZE]  ;
    }


    
    
  }
  /*** no pending anchor - revert to default ***/
  else {
    //  p->fighting_state += ( p->fighting_state - p->default_state ) % 2;
  }
  p->fighting_state += ( p->fighting_state < p->default_state ) ? 1 : -1;
  
  /*** v0.1.8 : give back stamina ***/
  if ( p->stamina < STAMINA_MAX ) { p->stamina++; }
  


  /* v0.2.3: different states, different speeds
   *
   *
   */
  switch ( running_state ) {
  case RUNSTATE_FIGHT:  timestep = 0.02;  break;
  case RUNSTATE_VICTORY:  timestep = 0.002;  break;
  default:  timestep = 0.002;  break;
  }


  /*** walls ***/
  if ( p->position[0] > 10.0 ) p->position[0] = 10.0;
  if ( p->position[2] > 10.0 ) p->position[2] = 10.0;
  if ( p->position[0] < -10.0 ) p->position[0] = -10.0;
  if ( p->position[2] < -10.0 ) p->position[2] = -10.0;

}


